
package jworms.game.weapons;

import jworms.game.GameField;
import jworms.game.GameMovingObject;
import jworms.game.GameObject;
import jworms.game.Player;
import jworms.game.explosion.Explosion;
import jworms.game.util.CollosionEnvironment;
import jworms.game.util.CollosionUtil;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public class Ball extends GameMovingObject implements Weapon{
// static
  /** ID for the object class. */
  private static int classID = 0;
  public int getClassID() { return classID; }
  public void setClassID(int id) { classID = id; }

  /**
   * Loads textures and sound.
   */
  public static void init(){
    BallUI.init();

    // register class id
    GameField.registerClass(new Ball(true, null, null, null));
  }

// non static
  /** GameField. */
  protected GameField gameField = null;
  /** Ball UI. */
  private BallUI ui = null;

  /** Position of the ball (in the game field). */
  protected Point2D.Float position = null;
  /** Movement speed/ direction vector. */
  protected Point2D.Float motion;
  /** Angle of the ball. */
  protected float angle = 0;
  /** Rotation of the ball. */
  protected float rotation = 0;

  /** Speed absorption when hit on ground. */
  protected float absobtion = 0.8f;
  /** Maximum motion speed. */
  private int MAX_MOTION = 15;

  /** Collosion vector. */
  private Point2D.Float hitVector = new Point2D.Float();

  /** Start time. */
  protected long startTime = 0;

  /**
   * Creates a ball.
   *
   * @param isHologram true if this is a hologram
   * @param gameField game field
   * @param position initial position of the ball
   */
  public Ball(boolean isHologram, GameField gameField, Point2D.Float position, Point2D.Float motion){
    super(isHologram);

    this.gameField = gameField;

    // set ball to the center of the game field
    this.position = position;
    this.motion = motion;

    // create ball UI
    ui = new BallUI(this, gameField);

    startTime = System.currentTimeMillis();
  }

  /**
   * Sets the ui of the ball.
   *
   * @param ui ui
   */
  public void setUI(BallUI ui){
    this.ui = ui;
  }

  /**
   * Returns the ui.
   *
   * @return ui
   */
  public BallUI getUI(){
    return ui;
  }

  public Point2D.Float getPosition(){
    return position;
  }

  public void doAction(){
  }

  /**
   * Lets the ball move.
   */
  public void move(){
    CollosionEnvironment collosion = ui.getCollosion();

    // go out of ground
    hitVector = CollosionUtil.getHitVector(position.x, position.y, collosion, gameField);
    float hitStrength = CollosionUtil.getLength(hitVector);

    // stop jumping too long
    float averageMotion = CollosionUtil.getLength(getLongAverageMotion());
    float absoluteAverageMotion = CollosionUtil.getLength(getAbsoluteLongAverageMotion());
    if(hitStrength > 0.5 && absoluteAverageMotion < 0.5 && (System.currentTimeMillis() - startTime) > 1000){
      motion.x *= 0.8;
      motion.y *= 0.8;
      rotation *= 0.95;

      if(absoluteAverageMotion < 0.3 && averageMotion < 0.05) return;
    }

    // calculate hit point with ground
    Rectangle2D.Float hitPoint = CollosionUtil.getHitPoint(position, motion, collosion, gameField);

    if(hitPoint != null){
      // hit vector
      hitVector.x = hitPoint.width;
      hitVector.y = hitPoint.height;

      // motion
      position.x = hitPoint.x;
      position.y = hitPoint.y;

      // hit on the ground
      Point2D.Float normal = CollosionUtil.getNormalized(hitVector);
      float hitNormalStrength = Math.abs(CollosionUtil.getScalarProduct(motion, normal));

      // rotation
      Point2D.Float parallelMotion =
              new Point2D.Float(motion.x + normal.x * hitNormalStrength, motion.y + normal.y * hitNormalStrength);
      rotation += CollosionUtil.getCrossProduct(parallelMotion, normal) / 90;

      // motion manipulation
      motion.x += 2 * normal.x * hitNormalStrength;
      motion.y += 2 * normal.y * hitNormalStrength;

      // absobtion
      motion.x *= absobtion;
      motion.y *= absobtion;

      // tell UI we hit on the ground
      ui.hitOnGround(absoluteAverageMotion);
    } else{
      // hit vector
      hitVector.x = 0;
      hitVector.y = 0;

      // ball motion
      position.x += motion.x;
      position.y += motion.y;
    }

    // gravity
    motion.y += GameField.GRAVITY;

    // go out of ground handling (very tricky)
    if(hitStrength > 1){
      float dx = hitVector.x / hitStrength;
      float dy = hitVector.y / hitStrength;

      if(averageMotion < 0.05){
        dx /= 20;
        dy /= 20;
      }

      position.x += dx;
      position.y += dy;

      dx /= 5;
      dy /= 5;
      if((dx > 0 && motion.x < dx) || (dx < 0 && motion.x > dx)) motion.x += dx;
      if((dy > 0 && motion.y < dy) || (dy < 0 && motion.y > dy)) motion.y += dy;
    }

    // motion slow down
    motion.x *= 0.995;
    motion.y *= 0.995;

    // rotation
    angle += rotation;
    if(angle < 0) angle += 2 * Math.PI;
    if(angle > 2 * Math.PI) angle -= 2 * Math.PI;
    rotation *= 0.995;

    // motion normalization
    motionNormalization();
  }

  /**
   * Draws the ball (calls BallUI.draw()).
   */
  public void draw(){
    ui.draw();
  }

  /**
   * Draws a symbol for the weapon.
   *
   * @param player player
   */
  public void drawSymbol(Player player){
    ui.drawSymbol(player);
  }

  public void dispose(){
    ui.dispose();
  }

  /**
   * Returns the angle of the ball.
   *
   * @return angle
   */
  public float getAngle(){
    return angle;
  }

  /**
   * Returns the motion angle.
   *
   * @return motion angle
   */
  public float getMotionAngle(){
    return (float) Math.atan2(-motion.y, motion.x);
  }

  /**
   * Keep motion in normal bounds.
   */
  private void motionNormalization(){
    float speed = (float) Math.sqrt(Math.pow(motion.x, 2) + Math.pow(motion.y, 2));

    if(speed > MAX_MOTION){
      motion.x *= MAX_MOTION / speed;
      motion.y *= MAX_MOTION / speed;
    }
  }

  /**
   * Returns the motion vector of the worm.
   *
   * @return motion vector
   */
  public Point2D.Float getMotion(){
    return motion;
  }

  /**
   * Sets the ball motion.
   *
   * @param motion motion
   */
  public void setMotion(Point2D.Float motion){
    this.motion = motion;
  }

// Weapon interface methods
  /** Loops till next shoot can be fired. */
  private int loopsTillReady = 0;

  /**
   * Created ball will not be able to move() or draw().
   */
  public Ball(GameField gameField){
    this(true, gameField, new Point2D.Float(), new Point2D.Float());
  }

  /**
   * Returns a shoot angle with a little bit random noise.
   *
   * @param player player
   * @return shoot angle
   */
  protected float getShootAngle(Player player){
    return player.getAimingAngle() + ((float) Math.random() - 0.5f) / 30;
  }

  public GameObject fireWeapon(Player player, GameField gameField){
    if(!isReady()) return null;

    // loops till next shoot
    loopsTillReady = 60;

    // calculate grenade
    float speed = player.getChargingValue() * 15;

    float angle = getShootAngle(player);
    float dx = (float) Math.cos(angle) * (player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1);
    float dy = (float) -Math.sin(angle);

    Point2D.Float position = new Point2D.Float(getCenter(player).x, getCenter(player).y);
    position.x += dx * 10;
    position.y += dy * 10;

    Point2D.Float motion = new Point2D.Float(dx * speed, dy * speed);

    return new Ball(false, gameField, position, motion);
  }

  public Point getCenter(Player player){
    return new Point((int) player.getInterpolatedPosition().x, (int) player.getInterpolatedPosition().y);
  }

  public boolean needAiming(){
    return true;
  }

  public boolean needsCharging(){
    return true;
  }

  public void loop(){
    if(loopsTillReady > 0) loopsTillReady--;
  }

  public boolean isReady(){
    return loopsTillReady == 0;
  }

  public float getReloadProcess(){
    return 1;
  }

// network
  public GameObject create(GameField gameField, DataInputStream in) throws IOException{
    Ball ball = new Ball(true, gameField, new Point2D.Float(), new Point2D.Float());
    ball.recieveUpdate(in);
    return ball;
  }

  public void sendUpdate(DataOutputStream out) throws IOException{
    super.sendUpdate(out);

    out.writeFloat(position.x);
    out.writeFloat(position.y);
    out.writeFloat(motion.x);
    out.writeFloat(motion.y);
    out.writeFloat(angle);
    out.writeFloat(rotation);
  }

  public void recieveUpdate(DataInputStream in) throws IOException{
    super.recieveUpdate(in);

    position.x = in.readFloat();
    position.y = in.readFloat();
    motion.x = in.readFloat();
    motion.y = in.readFloat();
    angle = in.readFloat();
    rotation = in.readFloat();
  }
}
