
package jworms.openGL;

import org.lwjgl.opengl.GL;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;
import java.util.ArrayList;

public class OpenGLTextureInternal{
// static properties
  /** Allowed texture sizes. */
  public static final int SIZES[] = {32, 64, 128, 256};

  /** ArrayList of textures to be created (OpenGLTextureInternal). */
  private static ArrayList genTextures = new ArrayList();

// OpenGL format constants
  /** RGB with 1 byte per color (3 bytes per pixel). */
  public static final int FORMAT_RGB = GL.GL_RGB;
  /** RGB with 5 bit per color (2 bytes per pixel). */
  public static final int FORMAT_RGB5 = GL.GL_RGB5;
  /** RGBA with 1 byte per color (4 bytes per pixel). */
  public static final int FORMAT_RGBA = GL.GL_RGBA;
  /** RGBA with 1 byte per color (4 bytes per pixel). */
  public static final int FORMAT_RGB5_A1 = GL.GL_RGB5_A1;
  /** RGB compressed. */
  public static final int FORMAT_COMPRESSED_RGB = GL.GL_COMPRESSED_RGB;

// texture properties
  /** Width of the texture. */
  public int width;
  /** Height of the texture. */
  public int height;

  /** OpenGL texture number. */
  public int nr;
  /** ByteBuffer with texture data. */
  public ByteBuffer buffer = null;

  /** Format of the buffer. */
  public int bufferFormat = 0;
  /** Format of the OpenGL texture. */
  public int internalFormat = 0;

  /** Should be set to true if the texture will never be modified. */
  public boolean isFinal = false;

  /** Filter for interpolation. */
  public int filter = GL.GL_NEAREST;

  /**
   * Creates a OpenGLTextureInternal with the given width and height.
   *
   * @param width width
   * @param height height
   * @param bufferFormat format of the buffer (FORMAT_...)
   * @param internalFormat format of the OpenGL texture (FORMAT_...)
   * @param isFinal true if the texture is final (will not be modified)
   */
  public OpenGLTextureInternal(int width, int height, int bufferFormat, int internalFormat, boolean isFinal){
    this.width = width;
    this.height = height;

    this.bufferFormat = bufferFormat;
    this.internalFormat = internalFormat;
    this.isFinal = isFinal;
  }

  /**
   * Loads the texture into OpenGL memory.
   */
  public void buffer2OpenGL(){
    if(buffer == null) return;

    // buffer 2 OpenGL
    GL.glBindTexture(GL.GL_TEXTURE_2D, nr);

    if(bufferFormat == FORMAT_COMPRESSED_RGB){
      // compressed texture
      GL.glCompressedTexImage2DARB(GL.GL_TEXTURE_2D,
                                   0,
                                   internalFormat,
                                   width,
                                   height,
                                   0,
                                   buffer.limit(),
                                   buffer);
    } else{
      // uncompressed texture
      GL.glTexImage2D(GL.GL_TEXTURE_2D,
                      0,
                      internalFormat,
                      width,
                      height,
                      0,
                      bufferFormat,
                      GL.GL_UNSIGNED_BYTE,
                      buffer);
    }
  }

  /**
   * Compresses the texture if possible.
   */
  public void compressTexture(){
    if(!isFinal || internalFormat != FORMAT_COMPRESSED_RGB || bufferFormat == internalFormat) return;

    // create one texture
    IntBuffer nrs = createIntBuffer(1);
    GL.glGenTextures(nrs);
    nr = nrs.get(0);

    // compress the texture
    buffer2OpenGL();

    // was the compression successful ?
    IntBuffer successBuffer = createIntBuffer(1);
    GL.glGetTexLevelParameter(GL.GL_TEXTURE_2D, 0, GL.GL_TEXTURE_COMPRESSED_ARB, successBuffer);

    if(successBuffer.get(0) != GL.GL_TRUE){
      throw new RuntimeException("Format GL_COMPRESSED_RGB not supported by your card.");
    }

    // format and size
    IntBuffer formatBuffer = createIntBuffer(1);
    GL.glGetTexLevelParameter(GL.GL_TEXTURE_2D, 0, GL.GL_TEXTURE_INTERNAL_FORMAT, formatBuffer);
    IntBuffer sizeBuffer = createIntBuffer(1);
    GL.glGetTexLevelParameter(GL.GL_TEXTURE_2D, 0, GL.GL_TEXTURE_COMPRESSED_IMAGE_SIZE, sizeBuffer);

    // read the compressed texture from OpenGL
    bufferFormat = FORMAT_COMPRESSED_RGB;
    internalFormat = formatBuffer.get(0);
    buffer = ByteBuffer.allocateDirect(sizeBuffer.get(0));
    GL.glGetCompressedTexImageARB(GL.GL_TEXTURE_2D, 0, buffer);

    // delete texture
    GL.glDeleteTextures(nrs);
  }

  /**
   * Adds the texture for later generation.
   *
   * @param texture texture
   */
  public static void addToGenTextures(OpenGLTextureInternal texture){
    genTextures.add(texture);
  }

  /**
   * Generates all loaded textures.
   * Must be called before textures can be used with OpenGL.
   */
  public static void glGenTextures(){
    // image addresses in memory
    IntBuffer nrs = createIntBuffer(genTextures.size());

    // create textures
    GL.glGenTextures(nrs);

    // bind the texture
    for(int texture_nr = 0; texture_nr < genTextures.size(); texture_nr++){
      OpenGLTextureInternal genTexture = (OpenGLTextureInternal) genTextures.get(texture_nr);
      genTexture.nr = nrs.get(texture_nr);

      GL.glBindTexture(GL.GL_TEXTURE_2D, genTexture.nr);
      GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, genTexture.filter);
      GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, genTexture.filter);
      GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE);
      GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE);

      genTexture.buffer2OpenGL();
    }
  }

  /**
   * Creates an IntBuffer of the given size.
   *
   * @param size size (in int)
   * @return IntBuffer
   */
  public static IntBuffer createIntBuffer(int size){
    return ByteBuffer.allocateDirect(4 * size).order(ByteOrder.nativeOrder()).asIntBuffer();
  }
}
